import os
import requests
import pathlib
from urllib.parse import urlparse
from typing import Generator, List
from cpacman.package import Package

class PackageManager():
	def __init__(self, clean_files:bool = False, *args, **kwargs) -> None:
		self.clean_files:bool = clean_files

	def __parse_requirements_line__(self, line:str, requirements_file:pathlib.Path) -> dict:
		# result is a dict of arguments
		# which will be givent to create new Package object
		result:dict = dict()

		# Splitting line into two parts
		# left part is a path where we will store a pacage
		# Right part sets url for downloading package
		#   author/packagename
		#   https://github.com/author/packagename
		#   https://github.com/author/packagename.git
		#   git@github.com:author/packagename.git
		#   git@github.com:author/packagename
		line:List[str] = line.split('>')

		if len(line) == 2:
			short_packagepath, link = line

			# Here we take info from a link
			# keys are:
			#   author       - author of the package
			#   projecktlink - link to github repo
			#   loadlink     - link to download repo in zip
			#   projecktname - name of the package
			link:dict = self.__parse_link__(link)
			reqfile_directory:pathlib.Path = pathlib.Path(requirements_file.resolve().parent)
			long_packagepath :pathlib.Path = reqfile_directory.joinpath(short_packagepath)

			# If boths varibles are not empty lines
			if link and long_packagepath:
				result['packagepath'] = long_packagepath
				result['loadlink'   ] = link['loadlink'   ]
				result['projectlink'] = link['projectlink']
				result['projectname'] = link['projectname']
		return result

	def __parse_link__(self, link:str) -> dict:
		# Here we take info from a link
		# keys are:
		#   author       - author of the package
		#   projecktlink - link to github repo
		#   loadlink     - link to download repo in zip
		#   projecktname - name of the package
		result:dict = dict()

		if link:
			url_path:str   = urlparse(link).path
			info:List[str] = url_path.split('/')

			# Left part is an author
			# Right part is project name
			if len(info) == 2:

				# Strip author and project
				info = list(map(lambda x: x.strip(), info))
				
				author, project = info

				if author and project:

					result['projectlink'] = f'https://github.com/{author}/{project}.git'
					result['loadlink'   ] = f'https://github.com/{author}/{project}/archive/master.zip'
					result['projectname'] = project
					result['author'     ] = author

		return result

	def read_requirements_file(self, requirements_file:pathlib.Path) -> Generator[Package, None, None]:
		with open(requirements_file, encoding="utf-8") as opened_requirements_file:
			for line in opened_requirements_file:
				info:Dict = self.__parse_requirements_line__(line, requirements_file)
				if info:
					yield Package(**info)
				else:
					print(f"Line was ignored due to incorrect schema!")
					print(f"    {line}")

	def load_packages_from_requirements_file(self, requirements_file:pathlib.Path):
		requirements_file:pathlib.Path = pathlib.Path(requirements_file)
		for package in self.read_requirements_file(requirements_file):
			package.load()
			package.unpack()
			if self.clean_files:
				package.clean()

	def install(self, line) -> None:
		info:dict = self.__parse_requirements_line__(line)
		if info:
			package:Package = Package(**info)
			package.load()
			package.unpack()
			if self.clean_files:
				package.clean()
		else:
			print(f"Line was ignored due to incorrect schema!")
			print(f"    {line}")

	def find_requirements_files(self):
		for filename in pathlib.Path.cwd().rglob("**/*"):
			if filename.is_file() and filename.suffix == '.cpacman':
				yield filename

	def recursive_install(self):
		for req_file in self.find_requirements_files():
			self.load_packages_from_requirements_file(req_file)
			



from setuptools import setup, find_packages

setup(
		name             = 'cpacman',
		version          = '1.0.4',
		packages         = find_packages(),
		author           = 'Miasnenko Dmitry',
		author_email     = 'cl0wzed.exe@gmail.com',
		url              = 'https://github.com/YoungMeatBoy/cpacman.git',
		install_requires = ['requests'],
		entry_points={
			'console_scripts': [
				'cpacman = cpacman.run:main'
		]
		}
)


import os
import shutil
import pathlib
import zipfile
import requests
from typing import List, Generator

class Package():
	def __init__(self,  packagepath:str, loadlink:str, projectlink:str, projectname:str, *args , **kwargs) -> None:
		# path to the place where it will be unpacked
		self.packagepath :pathlib.Path = pathlib.Path(packagepath)
		self.loadlink    :str          = loadlink
		self.projectlink :str          = projectlink
		self.zipfile_path:str          = pathlib.Path(str(packagepath).strip() + '.zip')
		self.projectname :str          = projectname
		self.packagepath :str          = self.patch_path(self.packagepath)
		self.is_loaded   :bool         = False
		self.is_unpacked :bool         = False
		self.is_cleaned  :bool         = False
		

	def load(self):
		print(f"Loading package: {self.projectlink}")
		response:requests.Response = requests.get(self.loadlink, stream=True)
		
		if response.status_code == 200:
			
			# Path where user wants to store the package: (self.packagepath)
			#   ./path/to/package                                   (1)
			# We create directory:
			#   ./path/to                                           (2)
			# Then we load zip file to this directory:    (self.zipfile_path)
			#   ./path/to/package.zip                               (3)
			# Then we unpack zip file to this directory and get folder:
			#   ./path/to/package-master/....(files of the package) (4)
			# Then we rename:
			#   ./path/to/package-master -> ./path/to/package       (5, 6)

			# (2)
			zipparent:pathlib.Path = pathlib.Path(self.zipfile_path.parent)
			zipparent.mkdir(parents=True, exist_ok=True)

			# (3)
			with open(self.zipfile_path, 'wb') as package_zip_file:
				for chunk in response.iter_content(chunk_size=1024):
					if chunk:
						package_zip_file.write(chunk)
			self.is_loaded = True
		else:
			print(f"Server returned bad status code: {response.status_code}")

	def unpack(self):
		if self.is_loaded:
			print(f"Unpacking: {self.zipfile_path}")
			# Move one step back
			outdir = self.packagepath.parent

			# (4)
			# Extract zip
			with zipfile.ZipFile(self.zipfile_path) as zip_ref:
				zip_ref.extractall(outdir)

			# Removing loaded zip file
			self.zipfile_path.unlink()
		
			# (5, 6)
			# Now it is saved in ./path/given/projectname-master
			# We rename to       self.packagepath
			previous_path = outdir.joinpath(f"{self.projectname}-master")
			
			if self.packagepath.exists():
				shutil.rmtree(self.packagepath)

			previous_path.rename(self.packagepath)
			
			self.is_unpacked = True
		else:
			print("Package was not loaded correctly!")
		
			
	def clean(self):
		if self.packagepath.exists() and self.is_loaded and self.is_unpacked:
			print(f"Cleaning directory: {self.unpatch_path(self.packagepath)}")
			extensions = ('.c', '.cpp', '.h', '.hpp', '.cpacman')
			for filename in self.packagepath.rglob('./**/*'):
				if filename.is_file() and not filename.suffix in extensions:
					print(f"    {self.unpatch_path(filename)}")
					filename.unlink()

	#   Sometimes search fails
	#   as system cannot find path
	#   This is a solution from StackOverflow
	#       https://stackoverflow.com/a/57502760
	@staticmethod
	def patch_path(path):
		normalized = os.fspath(path.resolve())
		if not normalized.startswith('\\\\?\\'):
			normalized = '\\\\?\\' + normalized.strip()
		return pathlib.Path(normalized)

	def unpatch_path(self, path):
		normalized = os.fspath(path.resolve())
		if normalized.startswith('\\\\?\\'):
			normalized = normalized.strip('\\\\?\\')
		return pathlib.Path(normalized)


def main():
    
	from cpacman.packagemanager import PackageManager
	from argparse import ArgumentParser

	parser = ArgumentParser()

	parser.add_argument('-i' , '--install'           , type = str)
	parser.add_argument('-r' , '--requirements'      , type = str)
	parser.add_argument('-c' , '--clean'             , action = 'store_true', default = False)
	parser.add_argument('-ri', '--recursive_install' , action = 'store_true', default = False)
	args = parser.parse_args()

	manager = PackageManager(clean_files = args.clean)

	if args.requirements:
		manager.load_packages_from_requirements_file(args.requirements)

	elif args.install:
		manager.install(args.install)

	elif args.recursive_install:
		manager.recursive_install()

if __name__ == '__main__':
    main()